import dataclasses
from datetime import date, datetime
import json
from pathlib import Path, PosixPath
from pydantic import BaseModel
from typing import Dict, List, Type, TypeVar

import yaml

T = TypeVar("T", bound=BaseModel)


def parse_dict_datamodel(
    obj_dict: Dict, 
    clazz: Type[T], 
) -> T:
    """
    Cast dict object to expected pydantic model
    :param obj_dict: dict object to be transformed to pydantic.BaseModel
    :param clazz: pydantic.BaseModel type
    """
    return clazz.parse_obj(obj_dict)


def parse_dict_datamodels(
    obj_list: List[Dict],
    clazz: Type[T],
) -> List[T]:
    """Cast list of dict objects to expected data model types (pydantic.BaseModel)"""
    return [parse_dict_datamodel(obj_dict, clazz) for obj_dict in obj_list]


def parse_yaml_datamodel(
    yaml_obj: str,
    clazz: Type[T],
) -> T:
    """
    Parse YAML and convert it to expected data model
    :param yaml_obj: YAML string
    :param clazz: pydantic.BaseModel type
    """
    data = yaml.load(yaml_obj, Loader=yaml.FullLoader)
    if data is None:
        data = {}
    return clazz.parse_obj(data)


def parse_yaml_file_datamodel(
    path: Path,
    clazz: Type[T],
) -> T:
    """
    Parse YAML file and convert it to expected data model
    :param path: Path to a YAML file
    :param clazz: pydantic.BaseModel type
    """
    assert path.is_file(), f"File doesn't exist: {path}"
    data = path.read_text()
    return parse_yaml_datamodel(data, clazz)


def datamodel_to_yaml_str(dt: BaseModel) -> str:
    data_dict = datamodel_to_dict(dt)
    return yaml.dump(data_dict)


def convert_to_json(obj) -> str:
    obj = convert_to_json_serializable(obj)
    obj = remove_none(obj)
    return json.dumps(obj)


def datamodel_to_dict(dt: BaseModel) -> Dict:
    data_dict = dt.dict()
    data_dict = remove_none(data_dict)
    data_dict = convert_to_json_serializable(data_dict)
    return data_dict


def remove_none(obj):
    """Remove unwanted null values"""
    if isinstance(obj, list):
        return [remove_none(x) for x in obj if x is not None]
    elif isinstance(obj, dict):
        return {k: remove_none(v) for k, v in obj.items() if k is not None and v is not None}
    else:
        return obj


def convert_to_json_serializable(obj):
    if dataclasses.is_dataclass(obj):
        return convert_to_json_serializable(dataclasses.asdict(obj))
    elif isinstance(obj, BaseModel):
        return obj.dict()
    elif isinstance(obj, PosixPath):
        return str(obj)
    elif isinstance(obj, (date, datetime)):
        return obj.isoformat()
    elif isinstance(obj, list):
        return [convert_to_json_serializable(x) for x in obj]
    elif isinstance(obj, dict):
        return {k: convert_to_json_serializable(v) for k, v in obj.items()}
    elif hasattr(obj, '__to_json__'):
        return getattr(obj, '__to_json__')()
    else:
        return obj



import re
from urllib.parse import urlsplit

from racetrack_client.log.context_error import wrap_context
from racetrack_client.manifest import Manifest
from racetrack_client.manifest.load import get_manifest_path
from racetrack_client.log.logs import get_logger
from racetrack_client.manifest.merge import load_merged_manifest
from racetrack_client.utils.datamodel import datamodel_to_yaml_str
from racetrack_client.utils.semver import SemanticVersion

logger = get_logger(__name__)


def load_validated_manifest(path: str) -> Manifest:
    """
    Load and validate manifest from a path. Raise exception in case of a defect.
    :param path path to a Job manifest file or to a directory with it
    :return loaded, valid Manifest
    """
    manifest_path = get_manifest_path(path)
    manifest = load_merged_manifest(manifest_path)

    with wrap_context('validating manifest'):
        validate_manifest(manifest)

    logger.debug(f'Manifest file "{manifest_path}" is valid')
    return manifest


def validate_manifest(manifest: Manifest):
    """Check whether manifest is valid. Raise exception in case of error"""
    assert re.match(r"[^@]+@[^@]+\.[^@]+", manifest.owner_email), '"owner_email" is not a valid email'

    with wrap_context('parsing Job version'):
        SemanticVersion(manifest.version)

    assert ':' in manifest.lang, '"lang" should specify the version in a format "name:version"'

    assert 1 <= manifest.replicas <= 15, 'replicas count out of allowed range'

    assert urlsplit(manifest.git.remote).scheme == 'https', 'git remote URL should be HTTPS'


def validate_and_show_manifest(path: str):
    manifest = load_validated_manifest(path)
    logger.info(f'Manifest file "{path}" is valid')
    print(datamodel_to_yaml_str(manifest))



import logging
import sys
from typing import Optional

LOG_FORMAT = '\033[2m[%(asctime)s]\033[0m %(levelname)s %(message)s'
LOG_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'


def init_logs():
    """Run logging initialization (before loading log level)"""
    logging.basicConfig(stream=sys.stdout, format=LOG_FORMAT, level=logging.INFO, datefmt=LOG_DATE_FORMAT)


def configure_logs(verbosity: int = 0, log_level: Optional[str] = None):
    """Configure root logger with a log level"""
    if log_level is None:
        if verbosity > 0:
            level = logging.DEBUG
        else:
            level = logging.INFO
    else:
        level = _get_logging_level(log_level)
    # Set root level to INFO to avoid printing a ton of garbage DEBUG logs from imported libraries
    # Proper log level is set on racetrack logger
    logging.basicConfig(stream=sys.stdout, format=LOG_FORMAT, level=logging.INFO, datefmt=LOG_DATE_FORMAT, force=True)

    for handler in logging.getLogger().handlers:
        handler.setFormatter(ColoredFormatter(handler.formatter))

    logger = logging.getLogger('racetrack')
    logger.setLevel(level)


def get_logger(logger_name: Optional[str] = None) -> logging.Logger:
    """Get configured racetrack logger"""
    logger = logging.getLogger('racetrack')
    if logger_name:
        return logger.getChild(logger_name)
    return logger


def _get_logging_level(str_level: str) -> int:
    return {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warn': logging.WARNING,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'crit': logging.CRITICAL,
        'off': logging.NOTSET,
    }[str_level.lower()]


class ColoredFormatter(logging.Formatter):
    def __init__(self, plain_formatter):
        logging.Formatter.__init__(self)
        self.plain_formatter = plain_formatter

    log_level_templates = {
        'CRITICAL': '\033[1;31mCRIT \033[0m',
        'ERROR': '\033[1;31mERROR\033[0m',
        'WARNING': '\033[0;33mWARN \033[0m',
        'INFO': '\033[0;34mINFO \033[0m',
        'DEBUG': '\033[0;32mDEBUG\033[0m',
    }

    def format(self, record: logging.LogRecord):
        if record.levelname in self.log_level_templates:
            record.levelname = self.log_level_templates[record.levelname].format(record.levelname)
        return self.plain_formatter.format(record)



from typing import Dict

from pydantic import BaseModel


class Credentials(BaseModel):
    username: str
    password: str


class ClientConfig(BaseModel, arbitrary_types_allowed=True):
    """Global options for a local client"""

    # default URL of Racetrack API server (Lifecycle URL)
    lifecycle_url: str = 'http://localhost:7002'

    # Git auth credentials set for particular repositories
    git_credentials: Dict[str, Credentials] = {}

    # Racetrack URL aliases: alias name -> full URL to Lifecycle API
    lifecycle_url_aliases: Dict[str, str] = {}

    # Auth tokens per Lifecycle URL
    user_auths: Dict[str, str] = {}



from contextlib import contextmanager

from racetrack_client.log.logs import get_logger

logger = get_logger(__name__)


class ContextError(RuntimeError):
    """
    Error wrapping other error, adding context message
    and printing it in a Go manner - from most general to the real cause,
    eg: 'initializing app: loading config: file is missing'

    Usage:
        try:
            raise RuntimeError('file is missing')
        except Exception as e:
            raise ContextError('loading config') from e
    """

    def __init__(self, context_message: str, cause: BaseException = None):
        super().__init__()
        self.context_message = context_message
        if cause is not None:
            self.__cause__ = cause

    def __str__(self):
        if self.__cause__ is None:
            return self.context_message
        else:
            return f'{self.context_message}: {self.__cause__}'


@contextmanager
def wrap_context(context_name: str, log_debug: bool = False):
    """Apply context to occurred errors and propagate them further"""
    if log_debug:
        logger.debug(context_name)
    try:
        yield
    except Exception as e:
        raise ContextError(context_name) from e



import os
from pathlib import Path
from typing import Type, TypeVar

from pydantic import BaseModel
import yaml

from racetrack_client.log.logs import get_logger

logger = get_logger(__name__)

T = TypeVar("T", bound=BaseModel)


def load_config(clazz: Type[T]) -> T:
    """
    Load general configuration from YAML file given in CONFIG_FILE environment var or load default config.
    :param clazz: pydantic.BaseModel type that should be loaded
    :return: configuration object of given "clazz" type
    """
    config_file_path = os.environ.get('CONFIG_FILE')
    if not config_file_path:
        logger.warning('CONFIG_FILE unspecified, loading default config')
        return clazz()

    path = Path(config_file_path)
    if not path.is_file():
        raise FileNotFoundError(f"config file {config_file_path} doesn't exist")

    try:
        with path.open() as file:
            config_dict = yaml.load(file, Loader=yaml.FullLoader)
            config = clazz.parse_obj(config_dict)

            logger.info(f'config loaded from {config_file_path}: {config}')
            return config
    except Exception as e:
        raise RuntimeError('loading config failed') from e



from enum import Enum
from typing import Dict, List, Optional

from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.auth import get_user_auth
from racetrack_client.client_config.io import load_client_config
from racetrack_client.log.logs import get_logger
from racetrack_client.utils.auth import get_auth_request_headers
from racetrack_client.utils.request import parse_response, parse_response_list, Requests
from racetrack_client.utils.table import print_table
from racetrack_client.utils.time import nullable_timestamp_pretty_ago


logger = get_logger(__name__)


class JobTableColumn(str, Enum):
    infrastructure = "infrastructure"
    deployed_by = "deployed_by"
    updated_at = "updated_at"
    last_call_at = "last_call_at"
    pub_url = "pub_url"
    all = "all"


def list_jobs(remote: Optional[str], columns: List[JobTableColumn]):
    """List all deployed jobs"""
    client_config = load_client_config()
    lifecycle_url = resolve_lifecycle_url(client_config, remote)
    user_auth = get_user_auth(client_config, lifecycle_url)

    r = Requests.get(
        f'{lifecycle_url}/api/v1/job',
        headers=get_auth_request_headers(user_auth),
    )
    jobs: List[Dict] = parse_response_list(r, 'Lifecycle response error')

    if not jobs:
        logger.info(f'No jobs deployed at {lifecycle_url}.')
        return

    header = ['NAME', 'VERSION', 'STATUS']
    if JobTableColumn.infrastructure in columns or JobTableColumn.all in columns:
        header.append('INFRASTRUCTURE')
    if JobTableColumn.deployed_by in columns or JobTableColumn.all in columns:
        header.append('DEPLOYED BY')
    if JobTableColumn.updated_at in columns or JobTableColumn.all in columns:
        header.append('UPDATED AT')
    if JobTableColumn.last_call_at in columns or JobTableColumn.all in columns:
        header.append('LAST CALL AT')
    if JobTableColumn.pub_url in columns or JobTableColumn.all in columns:
        header.append('PUB URL')

    table: List[List[str]] = [header]
    for job in jobs:
        name = job.get('name')
        version = job.get('version')
        status = job.get('status', '').upper()
        cells = [name, version, status]

        if JobTableColumn.infrastructure in columns or JobTableColumn.all in columns:
            cells.append(job.get('infrastructure_target'))
        if JobTableColumn.deployed_by in columns or JobTableColumn.all in columns:
            cells.append(job.get('deployed_by'))
        if JobTableColumn.updated_at in columns or JobTableColumn.all in columns:
            update_time = job.get('update_time')
            updated_ago = nullable_timestamp_pretty_ago(update_time)
            cells.append(updated_ago)
        if JobTableColumn.last_call_at in columns or JobTableColumn.all in columns:
            last_call_time = job.get('last_call_time')
            last_call_ago = nullable_timestamp_pretty_ago(last_call_time)
            cells.append(last_call_ago)
        if JobTableColumn.pub_url in columns or JobTableColumn.all in columns:
            cells.append(job.get('pub_url'))

        table.append(cells)

    logger.info(f'List of all jobs ({len(jobs)}) deployed at {lifecycle_url}:')
    print_table(table)


def move_job(remote: Optional[str], job_name: str, job_version: str, new_infra_target: str):
    """Move job from one infrastructure target to another"""
    client_config = load_client_config()
    lifecycle_url = resolve_lifecycle_url(client_config, remote)
    user_auth = get_user_auth(client_config, lifecycle_url)

    logger.info(f'Moving job "{job_name}" v{job_version} at {lifecycle_url} to {new_infra_target} infrastructure...')
    r = Requests.post(
        f'{lifecycle_url}/api/v1/job/{job_name}/{job_version}/move',
        json={
            'infrastructure_target': new_infra_target,
        },
        headers=get_auth_request_headers(user_auth),
    )
    parse_response(r, 'Lifecycle response error')
    logger.info(f'Job "{job_name}" v{job_version} at {lifecycle_url} has been moved to {new_infra_target} infrastructure ')


def delete_job(name: str, version: str, remote: Optional[str]):
    """Delete versioned job instance"""
    client_config = load_client_config()
    lifecycle_url = resolve_lifecycle_url(client_config, remote)
    user_auth = get_user_auth(client_config, lifecycle_url)

    logger.debug(f'Deleting job "{name}" v{version} from {lifecycle_url}...')
    r = Requests.delete(
        f'{lifecycle_url}/api/v1/job/{name}/{version}',
        headers=get_auth_request_headers(user_auth),
    )
    parse_response(r, 'Lifecycle response error')
    logger.info(f'Job "{name}" v{version} has been deleted from {lifecycle_url}')



from pathlib import Path
import stat
from typing import Dict, Optional

import yaml

from racetrack_client.log.context_error import wrap_context
from racetrack_client.client_config.client_config import ClientConfig, Credentials
from racetrack_client.log.logs import get_logger

logger = get_logger(__name__)


def load_client_config() -> ClientConfig:
    """Load global options for a local client"""
    path = Path.home() / '.racetrack' / 'config.yaml'
    if not path.is_file():
        return ClientConfig()
    
    _check_file_mode(path)

    try:
        with path.open() as file:
            config_dict = yaml.load(file, Loader=yaml.FullLoader)
            config = ClientConfig.parse_obj(config_dict)

            logger.debug(f'client config loaded from {path}')
            return config
    except Exception as e:
        raise RuntimeError('loading client config failed') from e


def load_credentials_from_dict(credentials_dict: Optional[Dict]) -> Optional[Credentials]:
    if credentials_dict is None:
        return None

    with wrap_context('parsing credentials'):
        return Credentials.parse_obj(credentials_dict)


def save_client_config(config: ClientConfig):
    data_dict = config.dict()

    dir_path = Path.home() / '.racetrack'
    dir_path.mkdir(parents=True, exist_ok=True)
    config_path = dir_path / 'config.yaml'
    if not config_path.is_file():
        config_path.touch(mode=0o600)

    yaml_content = yaml.dump(data_dict)
    config_path.write_text(yaml_content)


def _check_file_mode(config_path: Path):
    file_mode = config_path.stat().st_mode
    if file_mode & stat.S_IROTH:
        logger.warning(f'Config file {config_path} is readable by others! Change permissions to 600.')
    elif file_mode & stat.S_IRGRP:
        logger.warning(f'Config file {config_path} is readable by group! Change permissions to 600.')
    elif file_mode & stat.S_IWOTH:
        logger.warning(f'Config file {config_path} is writable by others! Change permissions to 600.')
    elif file_mode & stat.S_IWGRP:
        logger.warning(f'Config file {config_path} is writable by group! Change permissions to 600.')



from pathlib import Path
from typing import Dict

from racetrack_client.log.context_error import wrap_context
from racetrack_client.manifest import Manifest
from racetrack_client.manifest.load import load_manifest_dict_from_yaml, load_manifest_from_dict


def load_merged_manifest(manifest_path: Path) -> Manifest:
    """Load manifest from YAML file, resolve overlay layer, merging it with base manifest"""
    with wrap_context('loading manifest'):
        manifest_dict = load_manifest_dict_from_yaml(manifest_path)

    extends = manifest_dict.get('extends')
    if extends:
        extends_path = Path(extends)
        with wrap_context('extending base manifest with overlay'):
            with wrap_context('loading base manifest'):
                assert not extends_path.is_absolute(), 'extends path should be relative'
                base_path = manifest_path.parent / extends_path
                base_manifest_dict = load_manifest_dict_from_yaml(base_path)
                assert not base_manifest_dict.get('extends'), 'base manifest cannot extend another one'
            with wrap_context('merging base & overlay layers'):
                manifest_dict = merge_dicts(base_manifest_dict, manifest_dict)
                manifest_dict['extends'] = None

    return load_manifest_from_dict(manifest_dict)


def merge_dicts(base: Dict, overlay: Dict) -> Dict:
    """
    Merge dict overlay into base recursively (in case of same keys).
    Entries from overlay will overwrite entries from base.
    """
    for key in overlay:
        if key in base:
            if isinstance(base[key], dict) and isinstance(overlay[key], dict):
                base[key] = merge_dicts(base[key], overlay[key])
            else:
                base[key] = overlay[key]
        else:
            base[key] = overlay[key]
    return base



import os
import traceback
from typing import Collection, Iterable

from racetrack_client.log.logs import get_logger

logger = get_logger(__name__)


def short_exception_logger(exc_info, *_):
    """
    Display excpetion traceback in concise one-line format.
    Avoid printing long, superfluous lines of traceback, especially for multi-cause exceptions.
    """
    try:
        ex_type, e, tb = exc_info
        log_message = get_exc_info_details(ex_type, e, tb)
        logger.error(log_message)
    except BaseException as e:
        logger.exception(e)


def get_exc_info_details(ex_type, e, tb) -> str:
    """
    Return concise one-line details of exception, containing message, traceback and root cause type
    :param ex_type: Exception type
    :param e: Exception instance
    :param tb: Traceback object
    """
    traceback_ex = traceback.TracebackException(ex_type, e, tb, limit=None)
    traceback_lines = list(_get_traceback_lines(traceback_ex))
    traceback_str = ', '.join(traceback_lines)
    cause = _root_cause_type(e)
    return f'{str(e).strip()}, cause={cause}, traceback={traceback_str}'


def _root_cause_type(e: BaseException) -> str:
    while e.__cause__ is not None:
        e = e.__cause__
    return type(e).__name__


def log_exception(e: BaseException):
    """
    Log exception in concise one-line format containing message, exception type and short traceback
    """
    exc_info = (type(e), e, e.__traceback__)
    short_exception_logger(exc_info)


def short_exception_details(e: BaseException) -> str:
    """
    Return concise one-line details of exception, containing message, traceback and root cause type
    """
    return get_exc_info_details(type(e), e, e.__traceback__)


def _get_traceback_lines(t1: traceback.TracebackException) -> Iterable[str]:
    while True:
        frames: Collection[traceback.FrameSummary] = t1.stack
        for frame in frames:
            yield f'{os.path.normpath(frame.filename)}:{frame.lineno}'

        if t1.__cause__ is None:
            break
        t1 = t1.__cause__



from typing import List
import typer

from racetrack_client import __version__
from racetrack_client.client.deploy import BuildContextMethod, send_deploy_request, DeploymentError
from racetrack_client.client.manage import JobTableColumn, move_job, delete_job, list_jobs
from racetrack_client.client.logs import show_runtime_logs, show_build_logs
from racetrack_client.client_config.auth import login_user_auth, logout_user_auth
from racetrack_client.client_config.io import load_client_config
from racetrack_client.client_config.update import set_credentials, set_current_remote, get_current_remote, set_config_url_alias
from racetrack_client.plugin.bundler.bundle import bundle_plugin
from racetrack_client.plugin.install import install_plugin, list_available_job_types, list_installed_plugins, uninstall_plugin
from racetrack_client.client.run import run_job_locally
from racetrack_client.log.exception import log_exception
from racetrack_client.log.logs import configure_logs
from racetrack_client.log.logs import get_logger
from racetrack_client.manifest.validate import validate_and_show_manifest
from racetrack_client.utils.auth import AuthError
from racetrack_client.utils.datamodel import datamodel_to_yaml_str

logger = get_logger(__name__)

cli = typer.Typer(
    no_args_is_help=True,
    name='racetrack',
    help='CLI client tool for managing workloads in Racetrack',
)

def main():
    try:
        cli()
    except (DeploymentError, AuthError) as e:
        logger.error(str(e))  # no need for client's stacktrace in case of well known errors
    except Exception as e:
        log_exception(e)


@cli.callback()
def _startup(
    verbose: bool = typer.Option(False, '-v', '--verbose', help='enable verbose mode'),
):
    configure_logs(verbosity=1 if verbose else 0)


@cli.command('deploy')
def _deploy(
    workdir: str = typer.Argument(default='.', help='directory with job.yaml manifest'),
    _remote: str = typer.Argument(default=None, show_default=False, help="Racetrack server's URL or alias name (deprecated, use --remote)"),
    remote: str = typer.Option(default=None, show_default=False, help="Racetrack server's URL or alias name"),
    force: bool = typer.Option(False, '--force', help='overwrite existing job'),
    build_context: BuildContextMethod = typer.Option(BuildContextMethod.default, show_default=False, help='Force building job from local files ("local") or from git repository ("git")'),
):
    """Send request deploying a Job to the Racetrack cluster"""
    send_deploy_request(workdir, lifecycle_url=remote or _remote, force=force, build_context_method=build_context)


@cli.command('validate')
def _validate(
    path: str = typer.Argument(default='.', help='path to a Job manifest file or to a directory with it'),
):
    """Validate Job manifest file"""
    validate_and_show_manifest(path)


@cli.command('logs', no_args_is_help=True)
def _logs(
    name: str = typer.Argument(..., show_default=False, help='name of the job'),
    version: str = typer.Option('latest', show_default=True, help='version of the job'),
    remote: str = typer.Option(default=None, show_default=False, help="Racetrack server's URL or alias name"),
    tail: int = typer.Option(20, '--tail', help='number of recent lines to show'),
    follow: bool = typer.Option(False, '--follow', '-f', help='follow logs output stream'),
):
    """Show runtime logs from the output of a job"""
    show_runtime_logs(name, version, remote, tail, follow)


@cli.command('build-logs', no_args_is_help=True)
def _build_logs(
    name: str = typer.Argument(..., show_default=False, help='name of the job'),
    version: str = typer.Option('latest', show_default=True, help='version of the job'),
    remote: str = typer.Option(default=None, show_default=False, help="Racetrack server's URL or alias name"),
    tail: int = typer.Option(0, '--tail', help='number of recent lines to show, all logs by default'),
):
    """Show build logs from job image building"""
    show_build_logs(name, version, remote, tail)


@cli.command('list')
def _list_jobs(
    remote: str = typer.Option(default=None, show_default=False, help="Racetrack server's URL or alias name"),
    columns: List[JobTableColumn] = typer.Option([], '--column', '-c', show_default=False, help='Choose additional columns to show. "all" selects all columns.'),
):
    """List all deployed jobs"""
    list_jobs(remote, columns)


@cli.command('delete', no_args_is_help=True)
def _delete_job(
    name: str = typer.Argument(..., show_default=False, help='name of the job'),
    version: str = typer.Option(..., show_default=False, help='version of the job to delete'),
    remote: str = typer.Option(default=None, show_default=False, help="Racetrack server's URL or alias name"),
):
    """Delete job instance"""
    delete_job(name, version, remote)


@cli.command('move', no_args_is_help=True)
def _move_job(
    name: str = typer.Argument(..., show_default=False, help='name of the job'),
    version: str = typer.Option(..., show_default=False, help='version of the job to move out'),
    infrastructure: str = typer.Option(..., show_default=False, help='infrastructure target to move to'),
    remote: str = typer.Option(default=None, show_default=False, help="Racetrack server's URL or alias name"),
):
    """Move job from one infrastructure target to another"""
    move_job(remote, name, version, infrastructure)


@cli.command('run-local')
def _run_local(
    workdir: str = typer.Argument(default='.', help='directory with job.yaml manifest'),
    remote: str = typer.Option(default=None, show_default=False, help="Racetrack server's URL or alias name"),
    port: int = typer.Option(default=None, show_default=False, help='HTTP port to run the server on'),
    build_context: BuildContextMethod = typer.Option(BuildContextMethod.default, show_default=False, help='Force building job from local files ("local") or from git repository ("git")'),
):
    """Run job locally"""
    run_job_locally(workdir, remote, build_context_method=build_context, port=port)


@cli.command('version')
def _version():
    """Show the version information"""
    print(f'racetrack-client version {__version__}')


@cli.command('login', no_args_is_help=True)
def _login(
    user_token: str = typer.Argument(..., show_default=False, help='Racetrack Auth Token from Racetrack\'s user profile'),
    remote: str = typer.Option(default=None, show_default=False, help="Racetrack server's URL or alias name"),
):
    """Save user's Racetrack Auth Token for Racetrack server"""
    login_user_auth(remote, user_token)


@cli.command('logout')
def _logout(
    remote: str = typer.Option(default=None, show_default=False, help="Racetrack server's URL or alias name"),
):
    """Remove user's Racetrack Auth Token for Racetrack server"""
    logout_user_auth(remote)


# racetrack set ...
cli_set = typer.Typer(no_args_is_help=True, help='Set global options of the Racetrack client')
cli.add_typer(cli_set, name="set")

@cli_set.command('remote', no_args_is_help=True)
def _set_remote(
    remote: str = typer.Argument(..., show_default=False, help="Racetrack server's URL or alias name"),
):
    """Set current Racetrack's remote address"""
    set_current_remote(remote)


@cli_set.command('credentials', no_args_is_help=True)
def _set_credentials(
    repo_url: str = typer.Argument(..., show_default=False, help='URL of git remote for one of your jobs'),
    username: str = typer.Argument(..., show_default=False, help='username for git authentication'),
    token_password: str = typer.Argument(..., show_default=False, help='password or token for git authentication'),
):
    """Set read-access credentials for a git repository to build images from"""
    set_credentials(repo_url, username, token_password)


@cli_set.command('alias', no_args_is_help=True)
def _set_alias(
    alias: str = typer.Argument(..., show_default=False, help='short name for an environment'),
    racetrack_url: str = typer.Argument(..., show_default=False, help='URL address of a remote Racetrack server'),
):
    """Set up an alias for Racetrack's remote URL"""
    set_config_url_alias(alias, racetrack_url)


# racetrack get ...
cli_get = typer.Typer(no_args_is_help=True, help='Read global options of the Racetrack client')
cli.add_typer(cli_get, name="get")

@cli_get.command('remote')
def _get_remote():
    """Get current Racetrack's remote address"""
    get_current_remote()


@cli_get.command('config')
def _get_config():
    """Show all racetrack config values"""
    client_config = load_client_config()
    print(datamodel_to_yaml_str(client_config))


# racetrack plugin ...
cli_plugin = typer.Typer(no_args_is_help=True, help='Manage Racetrack plugins')
cli.add_typer(cli_plugin, name="plugin")

@cli_plugin.command('install', no_args_is_help=True)
def _install_plugin(
    plugin_uri: str = typer.Argument(..., show_default=False, help='location of the plugin file: local file path, HTTP URL to a remote file or repository name'),
    remote: str = typer.Option(default=None, show_default=False, help="Racetrack server's URL or alias name"),
):
    """Install a plugin to a remote Racetrack server"""
    install_plugin(plugin_uri, remote)


@cli_plugin.command('uninstall', no_args_is_help=True)
def _uninstall_plugin(
    plugin_name: str = typer.Argument(..., show_default=False, help='name of the plugin'),
    version: str = typer.Option(..., show_default=False, help='version of the plugin'),
    remote: str = typer.Option(default=None, show_default=False, help="Racetrack server's URL or alias name"),
):
    """Uninstall plugin from a remote Racetrack server"""
    uninstall_plugin(plugin_name, version, remote)


@cli_plugin.command('list')
def _list_installed_plugins(
    remote: str = typer.Option(default=None, show_default=False, help="Racetrack server's URL or alias name"),
    job_types: bool = typer.Option(False, '--job-types', help='list available job type versions'),
):
    """List plugins installed on a remote Racetrack server"""
    if job_types:
        list_available_job_types(remote)
    else:
        list_installed_plugins(remote)


@cli_plugin.command('bundle')
def _plugin_bundle(
    workdir: str = typer.Argument(default='.', help='path to a plugin directory'),
    out: str = typer.Option(default=None, show_default=False, help='output directory where to save the ZIP file'),
    plugin_version: str = typer.Option(default=None, show_default=False, help='override plugin version'),
):
    """Turn local plugin code into ZIP file"""
    bundle_plugin(workdir, out, plugin_version)



from typing import Optional, List, Dict, Any

from pydantic import BaseModel, Extra, validator

from racetrack_client.utils.quantity import Quantity


class PythonManifest(BaseModel, extra=Extra.forbid):
    requirements_path: Optional[str] = None
    # path to a Python file with a entrypoint class
    entrypoint_path: str = ''
    # base name of Python entrypoint class
    entrypoint_class: str = ''


class GoManifest(BaseModel, extra=Extra.forbid):
    # relative path to Go modules requirements
    gomod: str = 'go.mod'


class GitManifest(BaseModel, extra=Extra.forbid):
    # URL of git remote: HTTPS, SSH or directory path to a remote repository
    remote: str
    branch: Optional[str] = None
    # subdirectory relative to git repo root
    directory: str = '.'


class DockerManifest(BaseModel, extra=Extra.forbid):
    dockerfile_path: Optional[str] = None


class ResourcesManifest(BaseModel, extra=Extra.forbid, arbitrary_types_allowed=True):
    # minimum memory amount in bytes, eg. 256Mi
    memory_min: Optional[Quantity] = None
    # maximum memory amount in bytes, eg. 1Gi
    memory_max: Optional[Quantity] = None
    # minimum CPU consumption in cores, eg. 10m
    cpu_min: Optional[Quantity] = None
    # maximum CPU consumption in cores, eg. 1000m
    cpu_max: Optional[Quantity] = None


    @validator(
        'memory_min', 
        'memory_max', 
        'cpu_min',
        'cpu_max',
        pre=True)
    def _validate_quantity(cls, v):
        if v is None:
            return None
        return Quantity(str(v))


class Manifest(BaseModel, extra=Extra.forbid, arbitrary_types_allowed=True):
    """Job Manifest file - build recipe to get deployable image from source code workspace"""

    # name of the Job Workload
    name: str

    # email address of the Job's owner to reach out
    owner_email: str

    git: GitManifest

    # version of the Job
    version: str = '0.0.1'

    # Language wrapper used to embed model
    lang: str = 'python3'

    # relative path to base manifest file, which will be extended by this manifest
    extends: Optional[str] = None

    # Python-specific configuration
    python: Optional[PythonManifest] = None

    # Go/Golang-specific configuration
    golang: Optional[GoManifest] = None

    # Docker-specific configuration
    docker: Optional[DockerManifest] = None

    # type of deployed image: docker image, packer, AMI
    image_type: str = 'docker'

    # system-wide packages that should be installed with apt
    system_dependencies: Optional[List[str]] = None

    # env vars for building
    build_env: Optional[Dict[str, str]] = None
    # env vars for runtime
    runtime_env: Optional[Dict[str, str]] = None
    # secret env vars loaded from an external file applied on building
    secret_build_env_file: Optional[str] = None
    # secret env vars loaded from an external file applied at runtime
    secret_runtime_env_file: Optional[str] = None

    # labels - job metadata for humans
    labels: Optional[Dict[str, Any]] = None

    # list of public job endpoints that can be accessed without authentication
    public_endpoints: Optional[List[str]] = None

    # number of instances of the Job
    replicas: int = 1

    # resources demands to allocate to the Job
    resources: Optional[ResourcesManifest] = None

    # Language wrapper attributes
    wrapper_properties: Optional[Dict[str, Any]] = None

    # Back-end platform where to deploy the service
    infrastructure_target: Optional[str] = None



from pathlib import Path
from typing import Dict, Optional

import yaml
from racetrack_client.log.context_error import wrap_context
from racetrack_client.manifest.manifest import Manifest
from racetrack_client.utils.datamodel import parse_dict_datamodel, parse_yaml_datamodel
from racetrack_client.log.logs import get_logger

logger = get_logger(__name__)

JOB_MANIFEST_FILENAME = 'job.yaml'
FORMER_MANIFEST_FILENAME = 'fatman.yaml'

def load_manifest_from_yaml(path: Path) -> Manifest:
    """Load Manifest from a YAML file. Data types validation happens here when parsing YAML."""
    manifest_dict = load_manifest_dict_from_yaml(path)
    return load_manifest_from_dict(manifest_dict)


def load_manifest_dict_from_yaml(path: Path) -> Dict:
    if not path.is_file():
        raise FileNotFoundError(f"local manifest file '{path}' doesn't exist")

    with wrap_context('loading YAML'):
        with path.open() as file:
            return yaml.load(file, Loader=yaml.FullLoader)


def load_manifest_from_dict(manifest_dict: Dict) -> Manifest:
    """Return manifest as data class"""
    with wrap_context('parsing manifest data types'):
        return parse_dict_datamodel(manifest_dict, Manifest)


def parse_manifest_or_empty(manifest_yaml: Optional[str]) -> Optional[Manifest]:
    """Parse YAML string as Manifest. In case of error return None"""
    if not manifest_yaml:
        return None
    try:
        return parse_yaml_datamodel(manifest_yaml, Manifest)
    except Exception as e:
        logger.error(f'Job Manifest YAML contains syntax error: {e}')
        return None


def get_manifest_path(workdir_or_file: str) -> Path:
    if Path(workdir_or_file).is_dir():
        manifest_path = Path(workdir_or_file) / JOB_MANIFEST_FILENAME
        if manifest_path.is_file():
            return manifest_path
        
        former_path = Path(workdir_or_file) / FORMER_MANIFEST_FILENAME
        if former_path.is_file():
            logger.warning(f'You are using old manifest file name "{FORMER_MANIFEST_FILENAME}". Please rename it to "{JOB_MANIFEST_FILENAME}"')
            return former_path
        
        raise FileNotFoundError(f"local manifest file '{manifest_path}' doesn't exist")

    else:
        manifest_path = Path(workdir_or_file)
        if not manifest_path.is_file():
            raise FileNotFoundError(f"local manifest file '{manifest_path}' doesn't exist")

        if manifest_path.name == FORMER_MANIFEST_FILENAME:
            logger.warning(f'You are using old manifest file name "{FORMER_MANIFEST_FILENAME}". Please rename it to "{JOB_MANIFEST_FILENAME}"')
            
        return manifest_path



from functools import total_ordering
import re


@total_ordering
class Quantity:
    _suffix_multipliers = {
        'E': 1e18,
        'P': 1e15,
        'T': 1e12,
        'G': 1e9,
        'M': 1e6,
        'k': 1e3,
        '': 1,
        'm': 1e-3,
        'u': 1e-6,
        'n': 1e-9,
        'p': 1e-12,
        'Ei': 1024**6,
        'Pi': 1024**5,
        'Ti': 1024**4,
        'Gi': 1024**3,
        'Mi': 1024**2,
        'Ki': 1024,
    }

    def __init__(self, quantity_str: str):
        """
        Data type for representing fixed-point number with quantity suffixes
        Useful for measuring memory amount in bytes, or CPU consumption.

        :param quantity_str: plain integer, or fixed-point number
        using one of the quantity suffixes T, G, M, k, m, etc.
        You can also use the power-of-two equivalents: Ti, Gi, Mi, Ki
        For instance: 128974848, 129M, 128974848000m, 123Mi, 0.129G
        """
        self.quantity_str: str = quantity_str

        match = re.fullmatch(r'(?P<number>\d+(\.\d+)?)(?P<suffix>.*)', quantity_str)
        if match is None:
            raise ValueError(f'invalid quantity format: {quantity_str}, use "<number><suffix>"')

        self.base_number: float = float(match.group('number'))
        self.suffix: str = match.group('suffix')

        if not self.suffix in self._suffix_multipliers:
            raise ValueError(f'invalid suffix: {self.suffix}, use one of {self._suffix_multipliers.keys()}')

    def __str__(self) -> str:
        return self.quantity_str

    def __repr__(self) -> str:
        return self.__str__()
    
    def __bool__(self) -> bool:
        return self.base_number > 0

    def __eq__(self, other) -> bool:
        if not isinstance(other, Quantity):
            return NotImplemented
        return self.plain_number == other.plain_number

    def __lt__(self, other) -> bool:
        if not isinstance(other, Quantity):
            return NotImplemented
        return self.plain_number < other.plain_number

    def __to_json__(self) -> str:
        return self.quantity_str

    def __truediv__(self, other) -> 'Quantity':
        if isinstance(other, (int, float)):
            result_base_number = self.base_number / other
            if result_base_number.is_integer():
                result_base_number = int(result_base_number)
            return Quantity(f'{result_base_number}{self.suffix}')
        else:
            raise NotImplementedError()

    @property
    def plain_number(self) -> float:
        """Convert quantity to a plain number without any suffixes"""
        return self.base_number * self._suffix_multipliers[self.suffix]

    @classmethod
    def __get_validators__(cls):
        # Needed to make it compatible with pydantic's Custom Data Types
        # https://docs.pydantic.dev/usage/types/#custom-data-types
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if v is None:
            return None
        return Quantity(str(v))

    @classmethod
    def __modify_schema__(cls, field_schema):
        # Needed to make it compatible with pydantic's Custom Data Types
        field_schema.update(
            examples=["128974848", "129M", "128974848000m", "123Mi", "0.129G"],
        )



from typing import Optional

from pydantic import BaseModel, Extra


class PluginManifest(BaseModel, extra=Extra.forbid, arbitrary_types_allowed=True):
    """Details of the contents of the plugin"""

    name: str

    version: str

    build_date: Optional[str] = None
    
    # URL of the plugin homepage
    url: Optional[str] = None

    # order in plugins sequence, lowest priority gets executed first
    priority: int = 0

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))



import setuptools
from racetrack_client import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    install_requires = fh.read().splitlines()

setuptools.setup(
    name="racetrack-client",
    version=__version__,
    author='ERST',
    author_email="noreply@erst.dk",
    description="CLI client tool for deploying workloads to Racetrack",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TheRacetrack/racetrack",
    packages=setuptools.find_packages(exclude=["tests.*", "tests", "plugins"]),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Information Technology',
    ],
    python_requires='>=3.8.0',
    install_requires=install_requires,
    dependency_links=[],
    license='Apache License 2.0',
    entry_points={
        "console_scripts": [
            "racetrack = racetrack_client.main:main",
        ],
    },
)



from dataclasses import dataclass
import os
from pathlib import Path
import re
from typing import Dict, List, Optional, Tuple
from urllib.parse import urlparse

from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.auth import get_user_auth
from racetrack_client.client_config.client_config import ClientConfig
from racetrack_client.client_config.io import load_client_config
from racetrack_client.log.logs import get_logger
from racetrack_client.utils.auth import get_auth_request_headers
from racetrack_client.utils.request import parse_response, parse_response_object, parse_response_list, Requests
from racetrack_client.utils.semver import SemanticVersion

logger = get_logger(__name__)


def install_plugin(
    plugin_uri: str,
    lifecycle_url: Optional[str],
    client_config: Optional[ClientConfig] = None,
):
    """Install a plugin to a remote Racetrack server"""
    if client_config is None:
        client_config = load_client_config()
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    user_auth = get_user_auth(client_config, lifecycle_url)

    plugin_filename, plugin_content = _load_plugin_file(plugin_uri)

    logger.info(f'Uploading plugin {plugin_uri} to {lifecycle_url}')
    r = Requests.post(
        f'{lifecycle_url}/api/v1/plugin/upload/{plugin_filename}',
        data=plugin_content,
        headers=get_auth_request_headers(user_auth),
    )
    plugin_manifest = parse_response_object(r, 'Lifecycle response error')
    plugin_name = plugin_manifest.get('name')
    plugin_version = plugin_manifest.get('version')

    logger.info(f'Plugin {plugin_name} {plugin_version} has been installed to {lifecycle_url}')


def uninstall_plugin(
    plugin_name: str,
    plugin_version: str,
    lifecycle_url: Optional[str],
):
    """Uninstall plugin from a remote Racetrack server"""
    client_config = load_client_config()
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    user_auth = get_user_auth(client_config, lifecycle_url)

    logger.info(f'Uninstalling plugin {plugin_name} {plugin_version} from {lifecycle_url}')
    r = Requests.delete(
        f'{lifecycle_url}/api/v1/plugin/{plugin_name}/{plugin_version}',
        headers=get_auth_request_headers(user_auth),
    )
    parse_response(r, 'Lifecycle response error')

    logger.info(f'Plugin {plugin_name} {plugin_version} has been uninstalled from {lifecycle_url}')


def list_installed_plugins(lifecycle_url: Optional[str]) -> None:
    """List plugins installed on a remote Racetrack server"""
    client_config = load_client_config()
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    user_auth = get_user_auth(client_config, lifecycle_url)

    r = Requests.get(
        f'{lifecycle_url}/api/v1/plugin',
        headers=get_auth_request_headers(user_auth),
    )
    plugin_manifests: List[Dict] = parse_response_list(r, 'Lifecycle response error')
    plugin_infos = [f'{p["name"]}=={p["version"]}' for p in plugin_manifests]

    logger.info(f'Plugins currently installed on {lifecycle_url} ({len(plugin_infos)}):')
    for info in plugin_infos:
        print(info)


def list_available_job_types(lifecycle_url: Optional[str]) -> None:
    """List job type versions available on a remote Racetrack server"""
    client_config = load_client_config()
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    user_auth = get_user_auth(client_config, lifecycle_url)

    r = Requests.get(
        f'{lifecycle_url}/api/v1/plugin/job_type/versions',
        headers=get_auth_request_headers(user_auth),
    )
    versions: List[str] = parse_response_list(r, 'Lifecycle response error')

    logger.info(f'Job type versions currently installed on {lifecycle_url} ({len(versions)}):')
    for version in versions:
        print(version)


def _load_plugin_file(plugin_uri: str) -> Tuple[str, bytes]:
    local_file = Path(plugin_uri)
    if local_file.is_file():
        logger.debug(f'{plugin_uri} recognized as a local file')
        file_content = local_file.read_bytes()
        return local_file.name, file_content

    if plugin_uri.startswith('http') and plugin_uri.endswith('.zip'):
        logger.debug(f'{plugin_uri} recognized as a remote HTTP file')
        return download_file(plugin_uri)

    if plugin_uri.startswith('github.com/'):
        if '==' in plugin_uri:
            logger.debug(f'{plugin_uri} recognized as a specific version of a GitHub repository')
            parts = plugin_uri.split('==')
            return _get_github_release(parts[0], parts[1])
        else:
            logger.debug(f'{plugin_uri} recognized as a GitHub repository')
            return _get_latest_github_release(plugin_uri)

    raise ValueError(f'Unknown plugin location: {plugin_uri}')


def download_file(url: str) -> Tuple[str, bytes]:
    logger.info(f'downloading file {url}')
    filename = Path(urlparse(url).path).name
    response = Requests.get(url)
    response.raise_for_status()
    logger.debug(f'file {filename} downloaded ({len(response.content)} bytes)')
    return filename, response.content


@dataclass
class PluginRelease:
    version: str
    download_url: str


def _get_latest_github_release(repo_name: str) -> Tuple[str, bytes]:
    releases = _list_github_releases(repo_name)
    assert releases, f'repository {repo_name} doesn\'t have any valid release'
    latest_release = releases[-1]
    logger.info(f'getting latest version {latest_release.version} from the {repo_name} repository')
    return download_file(latest_release.download_url)


def _get_github_release(repo_name: str, version: str) -> Tuple[str, bytes]:
    releases = _list_github_releases(repo_name)
    assert releases, f'repository {repo_name} doesn\'t have any valid release'
    filtered_releases = [r for r in releases if r.version == version]
    assert filtered_releases, f'repository {repo_name} doesn\'t have a release with version {version}'
    release = filtered_releases[0]
    logger.info(f'getting version {release.version} from the {repo_name} repository')
    return download_file(release.download_url)


def _list_github_releases(repo_name: str) -> List[PluginRelease]:
    match = re.fullmatch(r'github.com/(.+)/(.+)', repo_name)
    assert match, f'can\'t match repo {repo_name} to a pattern'
    gh_user = match.group(1)
    gh_repo = match.group(2)

    headers = {
        'User-Agent': 'racetrack-client',
        'Accept': 'application/vnd.github+json',
        'X-GitHub-Api-Version': '2022-11-28',
    }
    if github_token := os.environ.get('GITHUB_TOKEN'):
        headers['Authorization'] = f'Bearer {github_token}'

    r = Requests.get(
        f'https://api.github.com/repos/{gh_user}/{gh_repo}/releases',
        headers=headers,
    )

    json_releases = parse_response_list(r, 'GitHub API response')
    releases = []
    for json_release in json_releases:
        version = json_release.get('tag_name')
        if not version:
            version = json_release.get('name')
        zip_assets = [asset['browser_download_url']
                      for asset in json_release['assets']
                      if asset['browser_download_url'].endswith('.zip')]
        if version and zip_assets:
            releases.append(PluginRelease(version, zip_assets[0]))

    releases.sort(key=lambda r: SemanticVersion(r.version))
    return releases



from enum import Enum
from typing import Dict

from racetrack_client.utils.request import parse_response_object, Requests


RT_AUTH_HEADER = 'X-Racetrack-Auth'


class RacetrackAuthMethod(Enum):
    TOKEN = 'racetrackAuth'


class AuthError(RuntimeError):
    def __init__(self, cause: str):
        super().__init__()
        self.cause = cause

    def __str__(self):
        return f'authentication error: {self.cause}'


def get_auth_request_headers(user_auth: str) -> Dict:
    return {
        RT_AUTH_HEADER: user_auth if user_auth != "" else None
    }


def is_auth_required(lifecycle_url: str) -> bool:
    r = Requests.get(
        f'{lifecycle_url}/api/v1/info',
    )
    response = parse_response_object(r, 'Lifecycle response error')
    return response["auth_required"]


def validate_user_auth(lifecycle_url: str, user_auth: str) -> str:
    r = Requests.get(
        f'{lifecycle_url}/api/v1/users/validate_user_auth',
        headers=get_auth_request_headers(user_auth),
    )
    if r.status_code == 401:
        raise AuthError('Invalid Auth Token')

    response = parse_response_object(r, 'Lifecycle response error')
    return response['username']
